%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This script was made for approximate measurement of AVR ADC bandwidth by 
%% means of amplitude spectrum comparison with reference signal sampled by 
%% wideband DSO directly at AVR ADC input pin.
%%  The script also includes basic measurement uncertainty analysis by means of 
%% Monte-Carlo method with support for multicore processing to make it 
%% reasonably fast.
%%  Script also contains correction of DSO timebase deviation from AVR clock.
%% This might be a bit funky because it uses leasqr() fitting of the waveform
%% by wave model (rectangular waveform). This little piece of code worked fine
%% for my DSO and AVR but it was quite sensitive to initial setup so it's good 
%% idea to verify its proper function.
%%  Input CSV files contains signed sample values organized in rows. Optimal
%% waveform for the script is square wave which contains enough harmonics to
%% compare.
%%
%% Version 2013.07.16 by Stanislav Maslan, s.maslan@seznam.cz
%% http://elektronika.kvalitne.cz/ATMEL/necoteorie/AVR_ETS_bandwidth/AVR_ETS_bandwidth.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% clear rubbish
clear all;
clc;

% switch plot output to GNUplot - works a little better under Windoze
graphics_toolkit('gnuplot');

%% Source data setup
mfld = mfilename('fullpath'); mfld = mfld(1:strchr(mfld,'\',1,'last'));
% data path (new Octave can't read data from '-path')
cd([mfld 'data']);
% source file postfix
pofx = '_wave.csv';
% column separator
clsep = ';';

%% AVR wave setup
% sampling rate
avr_sr = 20e+6;
% periods count
avr_pn = 16;
% file name prefix
AVR_prfx = 'AVR';

%% DSO wave setup
% sampling rate
dso_sr = 500e+6;
% limit samples count
dso_maxm = 200000;
% file nam prefix
DSO_prfx = 'DSO';

%% limit analysis freq. [Hz]
f_lim = 2e6;



%% read AVR wave
printf('Reading AVR wave ... ');
% read the wave
avr = dlmread([AVR_prfx pofx],clsep);
n = length(avr);
% calc. wave fundamental freq.
sig_f = avr_sr/n*avr_pn;
printf('done (%d samples)\n',n);

%% read DSO wave
printf('Reading DSO wave ... ');
% read wave
csv = dlmread([DSO_prfx pofx],clsep,[0 0 dso_maxm-1 0]);
% downsample (just for display - not used for calculations)
dso = decimate(csv,dso_sr/avr_sr,'FIR');
m = length(dso);
printf('done (%d samples)\n',length(csv));

%% find best correlation of AVR/DSO waveforms
np = n/avr_pn; 
avrp = avr(1:np);
[xc] = xcorr(dso,avrp);
xc = xc(m:end);
[v,id] = max(xc);
if(id+n>m)
    id = id - np;
end;
id = id + 0;

% take one chunk
sdso = dso(id:id+np-1);

%% match amplitudes
sdso = sdso*sumsq(avrp).^0.5/sumsq(sdso).^0.5;

% display waves
figure(1);
subplot(2,1,1);
plot(0:np-1,sdso);
hold on;
plot(0:np-1,avrp,'r');
hold off;
ylabel('y [-]','FontWeight','bold');
title(['Waves ' num2str(0.001*sig_f,4) 'kHz'],'FontWeight','bold');
legend('DSO','AVR','Location','NorthEast');




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% And now serious stuff %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% find frequency deviation of DSO timebase from AVR XTAL %%%
printf('Correcting DSO time base deviation ... ');

% DSO time scale
m = length(csv);
t = (0:(m-1))/dso_sr;

% filter data and make square wave of it
fil_n = 10;
csvf = sign(conv(csv,ones(fil_n,1),'same'));
csvf = csvf(fil_n:end-fil_n);
tf = t(fil_n:end-fil_n);

%% coarse fit with sine model 
% initial wave model parameters
th = [sig_f 0 0 max(csvf)];
% create wave model
FW = inline(" th(3)*sin((t+th(2))*2*pi*th(1)) + th(4) ","t","th");  
% fit DSO wave with the model
[rfcomp,rp,rkvg,riter,rcorp,rcovp,rcovr,rstdresid,rZ,rr2]=leasqr(tf,csvf',th,FW,0.001,10000);

%% and fine fit with variable duty-cycle square wave
% initial wave model parameters
th = [rp(1:2)' 0.5];
% create wave model
FW = inline(" 2*(double)(mod(tf+th(2),1/th(1))<(th(3)/th(1)))-1 ","t","th");  
% fit DSO wave with the model
[rfcomp,rp,rkvg,riter,rcorp,rcovp,rcovr,rstdresid,rZ,rr2]=leasqr(tf,csvf',th,FW,0.001,10000);

% time scale correction 
tsr = sig_f/rp(1);


printf('done (%+dppm)\n',(rp(1)-sig_f)/sig_f*1e6);



%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Monte-carlo setup %%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%

str = 'Monte-Carlo uncertainty analysis';
printf([repmat('-',1,length(str)) '\n' str '\n' repmat('-',1,length(str)) '\n']);

% MC cycles
mc = 20000;
% MC multithreading cores
mc_cores = 4;

%% source uncertainties, all values are expected in gaussian distribution
% ref. freq uncertainty [dB]
par.uc_rf = 0.5;
% DSO flatness uncertainty [dB]
par.uc_d_flat = 0.3;
% DSO relative ADC quantization noise [-]
par.uc_d_noise = 8/256;
% DSO jitter noise [s]
par.uc_d_jt = 5e-10;
% DSO time base instability relative to AVR [1/s]
par.uc_d_tstab = 100e-6;
% DSO non-linearity
par.uc_d_inl = 4/256;
% DSO non-linearity order
par.uc_d_inl_p = 4;
% AVR relative ADC quantization noise [-]
par.uc_a_noise = 8/1024;
% AVR non-linearity
par.uc_a_inl = 4/1024;
% AVR non-linearity order
par.uc_a_inl_p = 4;

% other paramters for MC
par.avr = avr;
par.avr_sr = avr_sr;
par.avr_pn = avr_pn;
par.n = n;
par.sig_f = sig_f;
par.t = t;
par.csv = csv;
par.dso_sr = dso_sr;
par.tsr = tsr;
par.f_lim = f_lim;

%% main MC processing
tic();
% generate job share folder path
mcfld = mfilename('fullpath');
mcfld = [mfld 'mc_rubbish'];
% process MC
res = multicore_mc_chunks(@AVR_bw_mc,par,mc,20,200,mcfld,min(mc,mc_cores));

% get results
avr_f = res.avr_f(1,:);
a_lst = res.a;
f0_lst = res.f0;

printf(strftime('Elapsed time: %H:%M:%S\n',gmtime(toc())));
printf([repmat('-',1,length(str)) '\n']);

% MC mean result
if(mc==1)
  a = a_lst;  
else 
  a = mean(a_lst);
end
f0 = mean(f0_lst);
% MC uncertainty
if(mc==1)
  uc_a = zeros(length(a),1);
  uc_f0 = 0;
else
  uc_a = std(a_lst);
  uc_f0 = std(f0_lst);
end

%% add zero freq.
a = [0 a];
uc_a = [uc_a(1) uc_a];
avr_f = [1e4 avr_f];

printf('%dkHz bandwidth for -3dB\n',0.001*f0);



%% plot frequency transfer
subplot(2,1,2);
semilogx(avr_f,a,'LineWidth',2);
grid on;
hold on;
semilogx(avr_f,a+uc_a,'b--');
semilogx(avr_f,a-uc_a,'b--');
v = axis;
%v(3) = -15;
v(2) = f_lim;
semilogx([f0 f0],[-3 v(3)],'r','LineWidth',2);
semilogx([v(1) f0],[-3 -3],'r','LineWidth',2);
semilogx([f0+uc_f0 f0+uc_f0],[-3 v(3)],'r--');
semilogx([v(1) f0+uc_f0],[-3 -3],'r--');
semilogx([f0-uc_f0 f0-uc_f0],[-3 v(3)],'r--');
semilogx([v(1) f0-uc_f0],[-3 -3],'r--');
axis(v);
hold off;
xlabel('f [Hz]','FontWeight','bold');
ylabel('A [dB]','FontWeight','bold');
title(['Frequency transfer, f = (' num2str(1e-3*f0,3) ' \pm ' num2str(1e-3*uc_f0,2) ') kHz at -3 dB'],'FontWeight','bold');
% export image
print('res_graph.png','-S640,480','-F:7');

%% plot MC histogram
figure(2);
hist(1e-3*f0_lst,30);
xlabel('f [kHz]','FontWeight','bold');
ylabel('N [-]','FontWeight','bold');
title('Monte-Carlo distribution of -3dB frequency','FontWeight','bold');
% export image
print('res_hist.png','-S640,240','-F:7');